package cn.iocoder.yudao.module.remote.api.robot;

import cn.hutool.core.util.ObjectUtil;
import cn.iocoder.yudao.module.remote.controller.admin.robot.dto.RemoteControllerSocketDTO;
import cn.iocoder.yudao.module.remote.enums.robot.RemoteIpTypeEnum;
import cn.iocoder.yudao.module.remote.util.crc.CRCUtil;
import cn.iocoder.yudao.module.system.api.remote.dto.RemoteRobotTransferDTO;
import com.alibaba.fastjson.JSON;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.PreDestroy;
import java.io.IOException;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.system.enums.ErrorCodeConstants.REMOTE_ROBOT_CONNECT_FAIL;

@Component
@Slf4j
public class RemoteControllerProcessor {
    private final ConcurrentHashMap<String, RemoteControllerSocketDTO> cache = new ConcurrentHashMap<>();

    private final ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(1);

    @Value("${remote.msg}")
    private String defaultMsg;

    @Value("${remote.controller-ip}")
    private String remoteControllerIp;

    @Value("${remote.controller-port}")
    private int remoteControllerPort;

    @Value("${remote.cockpit-time-out}")
    private int cockpitTimeOut;

    @Value("${remote.industrial-control-time-out}")
    private int industrialControlTimeOut;

    @Value("${remote.robot-communication-time-out}")
    private int robotCommunicationTimeOut;

    private final String ONE = "01";
    private final String TWO = "02";

    public void addCache(RemoteRobotTransferDTO dto) {
        RemoteControllerSocketDTO remoteControllerSocketDTO = cache.get(remoteControllerIp);
        log.info("缓存的数据 :{}", JSON.toJSONString(remoteControllerSocketDTO));
        if (ObjectUtil.isNotEmpty(remoteControllerSocketDTO)) {
            try {
                Socket socket = remoteControllerSocketDTO.getSocket();
                if (socket != null && !socket.isClosed()) {
                    socket.close();
                }
                log.info("关闭socket :{}", remoteControllerIp);
            } catch (IOException e) {
                log.error("关闭socket出现异常 :{}", remoteControllerIp);
                throw new RuntimeException(e);
            }
        } else {
            remoteControllerSocketDTO = new RemoteControllerSocketDTO();
        }
        remoteControllerSocketDTO.setHost(remoteControllerIp);
        remoteControllerSocketDTO.setPort(remoteControllerPort);
        Socket socket = new Socket();
        try {
            socket.connect(new InetSocketAddress(remoteControllerIp, remoteControllerPort), 1000);
            remoteControllerSocketDTO.setSocket(socket);
        } catch (IOException e) {
            log.error("添加socket失败 :{}", e);
            throw exception(REMOTE_ROBOT_CONNECT_FAIL);
        }
        setMsg(remoteControllerSocketDTO, RemoteIpTypeEnum.ONE.getType(), dto);
        cache.put(remoteControllerIp, remoteControllerSocketDTO);
    }

    /**
     * 设置
     *
     * @param remoteControllerSocket
     * @param type
     */
    public void setMsg(RemoteControllerSocketDTO remoteControllerSocket, Integer type, RemoteRobotTransferDTO dto) {
        String msg = defaultMsg;
        if (RemoteIpTypeEnum.ONE.getType().equals(type)) {
            msg = msg + " " + ONE;
        } else {
            msg = msg + " " + TWO;
        }

        String[] split = dto.getRobotIp().split("\\.");
        for (String ipItem : split) {
            String hex = Integer.toHexString(Integer.valueOf(ipItem));
            msg = msg + " " + hex;
            log.info("组装 :{}", hex);
        }

        String portHex = Integer.toHexString(dto.getRobotPort().intValue());
        String portHexOne = portHex.substring(0, 2);
        String portHexTwo = portHex.substring(2);
        msg = msg + " " + portHexOne + " " + portHexTwo;
        if (cockpitTimeOut > 0) {
            msg = msg + " " + ONE + " " + Integer.toHexString(cockpitTimeOut);
        } else {
            msg = msg + " " + TWO + " " + Integer.toHexString(cockpitTimeOut);
        }

        if (industrialControlTimeOut > 0) {
            msg = msg + " " + ONE + " " + Integer.toHexString(industrialControlTimeOut);
        } else {
            msg = msg + " " + TWO + " " + Integer.toHexString(industrialControlTimeOut);
        }

        if (robotCommunicationTimeOut > 0) {
            msg = msg + " " + ONE + " " + Integer.toHexString(robotCommunicationTimeOut);
        } else {
            msg = msg + " " + TWO + " " + Integer.toHexString(robotCommunicationTimeOut);
        }

        String crc = CRCUtil.getCRC(msg);
        String crcOne = crc.substring(0, 2);
        String crcTwo = crc.substring(2);

        msg = msg + " " + crcOne + " " + crcTwo;

        remoteControllerSocket.setMsg(msg);
    }

    public void remoteCache(RemoteRobotTransferDTO dto) {
        RemoteControllerSocketDTO remoteControllerSocketDTO = cache.get(remoteControllerIp);
        if (ObjectUtil.isEmpty(remoteControllerSocketDTO)) {
            return;
        }
        setMsg(remoteControllerSocketDTO, RemoteIpTypeEnum.THREE.getType(), dto);
        cache.remove(remoteControllerIp);

        Socket socket = remoteControllerSocketDTO.getSocket();
        if (socket == null ||socket.isClosed()) {
            try {
                socket = new Socket();
                socket.connect(new InetSocketAddress(remoteControllerIp, remoteControllerPort), 1000);
            } catch (IOException e) {
                log.error("连接socket出现异常 :{}", remoteControllerIp);
            }
        }

        OutputStream os = null;
        try {
            os = socket.getOutputStream();
            log.info("断开连接 :{} ,对应的IP  :{}", remoteControllerSocketDTO.getMsg(), remoteControllerSocketDTO.getHost());
            os.write(remoteControllerSocketDTO.getMsg().getBytes());
            socket.shutdownInput();
            socket.close();
            log.info("关闭socket :{}", remoteControllerIp);
        } catch (IOException e) {
            log.error("关闭socket出现异常 :{}", remoteControllerIp);
        } finally {
            if (ObjectUtil.isNotEmpty(os)) {
                try {
                    os.close();
                } catch (IOException e) {
                }
            }
        }

        remoteCache(dto);
    }


    public RemoteControllerProcessor() {
        // 每秒执行一次 - 处理并发送数据 - 避免数据丢失
        scheduler.scheduleAtFixedRate(this::processAndSend, 150, 150, TimeUnit.MILLISECONDS);
    }

    private void processAndSend() {
        if (ObjectUtil.isEmpty(cache)) {
            return;
        }
        log.info("socket发送数据开始");
        for (Map.Entry<String, RemoteControllerSocketDTO> v : cache.entrySet()) {
            RemoteControllerSocketDTO remoteControllerSocketDTO = v.getValue();
            if (ObjectUtil.isEmpty(remoteControllerSocketDTO)) {
                remoteControllerSocketDTO = new RemoteControllerSocketDTO();
                remoteControllerSocketDTO.setHost(v.getKey());
                remoteControllerSocketDTO.setPort(remoteControllerPort);
                try {
                    Socket socket = new Socket();
                    socket.connect(new InetSocketAddress(v.getKey(), remoteControllerPort), 1000);
                    remoteControllerSocketDTO.setSocket(socket);
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
                cache.put(v.getKey(), remoteControllerSocketDTO);
            }

            Socket socket = remoteControllerSocketDTO.getSocket();
            if (socket == null || socket.isClosed()) {
                socket = new Socket();
                try {
                    socket.connect(new InetSocketAddress(v.getKey(), remoteControllerPort), 1000);
                    remoteControllerSocketDTO.setSocket(socket);
                } catch (IOException e) {
                    log.error("socket重连异常 :{}", e);
                }
            }

            OutputStream os = null;
            try {
                os = socket.getOutputStream();
                String str = remoteControllerSocketDTO.getMsg();
                log.info("socket推送的数据 :{}", str);
                os.write(remoteControllerSocketDTO.getMsg().getBytes());
            } catch (IOException e) {
                log.error("socket发送异常 :{}", e);
            } finally {
                if (ObjectUtil.isNotEmpty(os)) {
                    try {
                        os.close();
                    } catch (IOException e) {
                    }
                }
            }
            log.info("socket发送数据成功");
        }
    }


    public void shutdown() {
        scheduler.shutdown();
    }

    @PreDestroy
    public void destroy() {
        // 在Bean销毁前关闭去啊
        shutdown();
    }
}
